load("NativeMemoryOperations.js");
NativeMemoryOperations.installPolyfills();

function simpleMemcpy(dest, destOffset, src, srcOffset, countBytes) {
  var destBytes = new Uint8Array(dest.buffer, 0, dest.buffer.byteLength);
  var srcBytes = new Uint8Array(src.buffer, 0, src.buffer.byteLength);
  destOffset = ((destOffset * dest.BYTES_PER_ELEMENT) + dest.byteOffset) | 0;
  srcOffset = ((srcOffset * src.BYTES_PER_ELEMENT) + src.byteOffset) | 0;
  var destEndOffset = (destOffset + countBytes) | 0;
  for (; destOffset < destEndOffset; destOffset = (destOffset + 1) | 0, srcOffset = (srcOffset + 1) | 0)
    destBytes[destOffset] = srcBytes[srcOffset];
}

function simpleMemcpySlow(dest, destOffset, src, srcOffset, countBytes) {
  var destBytes = new Uint8Array(dest.buffer, 0, dest.buffer.byteLength);
  var srcBytes = new Uint8Array(src.buffer, 0, src.buffer.byteLength);
  destOffset = ((destOffset * dest.BYTES_PER_ELEMENT) + dest.byteOffset) | 0;
  srcOffset = ((srcOffset * src.BYTES_PER_ELEMENT) + src.byteOffset) | 0;
  
  for (var i = 0; i < countBytes; ++i)
    destBytes[destOffset++] = srcBytes[srcOffset++];
}

function typedArraySetMemcpy(dest, destOffset, src, srcOffset, countBytes) {
  destOffset = ((destOffset * dest.BYTES_PER_ELEMENT) + dest.byteOffset) | 0;
  srcOffset = ((srcOffset * src.BYTES_PER_ELEMENT) + src.byteOffset) | 0;

  var destBytes = new Uint8Array(dest.buffer, 0, dest.buffer.byteLength);
  var srcBytes = new Uint8Array(src.buffer, srcOffset, countBytes);

  destBytes.set(srcBytes, destOffset);
}

function arraysAreEqual(a, b) {
  if (a.length !== b.length)
    return false;
  for (var i = 0, l = a.length; i < l; i++)
    if (a[i] !== b[i])
      return false;
  return true;
}

var bufferSize = 4096;

var sourceArray = new Uint32Array(bufferSize);
var destArray = new Uint32Array(bufferSize);

var warmingIterations = 32;
var iterations = 256;

function timeFunction (fn) {
  putstr(fn.name + ": ");
  
  for (var i = 0; i < warmingIterations; i++)
    fn();
  
  var started = Date.now();
  try {
    for (var i = 0; i < iterations; i++) {
      fn();
    }
  } finally {
  }
  var ended = Date.now();
  
  print((ended - started).toFixed(3));
};

function benchNativeMemoryOperations () {
  for (var i = 0; i < iterations; i++)
    NativeMemoryOperations.memcpy(destArray, 0, sourceArray, 0, sourceArray.byteLength);

  if (!arraysAreEqual(destArray, sourceArray))
    throw new Error("failed");
};

function benchSimpleMemcpySlow () {
  for (var i = 0; i < iterations; i++)
    simpleMemcpySlow(destArray, 0, sourceArray, 0, sourceArray.byteLength);

  if (!arraysAreEqual(destArray, sourceArray))
    throw new Error("failed");
};

function benchSimpleMemcpy () {
  for (var i = 0; i < iterations; i++)
    simpleMemcpy(destArray, 0, sourceArray, 0, sourceArray.byteLength);

  if (!arraysAreEqual(destArray, sourceArray))
    throw new Error("failed");
};

function benchTypedArraySet () {
  for (var i = 0; i < iterations; i++)
    typedArraySetMemcpy(destArray, 0, sourceArray, 0, sourceArray.byteLength);

  if (!arraysAreEqual(destArray, sourceArray))
    throw new Error("failed");
};

function runBenchmark () {
  for (var i = 0; i < bufferSize; i++)
    sourceArray[i] = i;
  
  timeFunction(benchTypedArraySet);
  timeFunction(benchNativeMemoryOperations);
  timeFunction(benchSimpleMemcpySlow);
  timeFunction(benchSimpleMemcpy);
};

runBenchmark();